#include "floatdefs.hh"
#include "matrix.hh"
#include "xyz.hh"
#include "map.hh"
#include "rgb.hh"
#include "trace.hh"
#include "object.hh"

#include <stdio.h>

class Light
{
public:
    XYZ pos;
    RGB type;
    FLOAT rad;

    Light(const XYZ &p, const RGB &c, FLOAT_ARG r) : pos(p), type(c), rad(r)
    {
    }
};

#include <list>
class Lights : public std::list<Light>
{
public:
    Lights()
    {
    #if 0
        for(unsigned a=0; a<3; ++a)
        {
            FLOAT r=drand48(), g=drand48(), b=drand48();
            FLOAT pow = max(max(r,g),b);
            r /= pow; g /= pow; b /= pow;
            
            push_back(Light(
                XYZ(drand48()*240+130,
                    drand48()*240+130,
                    drand48()*150+25),
                RGB(r,g,b),
                2));
        }
    #endif
    #if 0
        push_back(Light(XYZ(10,10,10), RGB(1,1,1), 4));
        push_back(Light(XYZ(100,480,100), RGB(1.4,2,2), 4));
    #endif
    #if 0
        push_back(Light(XYZ(-1200,  1700,440), RGB(1,1,1), 4));
        push_back(Light(XYZ(-1200, -1200,440), RGB(1,1,1), 4));
        push_back(Light(XYZ( 1700, -1200,440), RGB(1,1,1), 4));
        push_back(Light(XYZ( 1700,  1700,440), RGB(1,1,1), 4));
    #endif
        
        push_back(Light(XYZ( 70,60,180), RGB(1,1,1), 4));
    }
};
static Lights lights;

#include "redirect.hh"

static unsigned recdepth=0;

/*
    Trace(m, eye, dir, effect, ior)

       m is the map that describes the world.
       eye is the starting point of the beam
       dir is the direction the beam runs to
       effect is a number that tells how much effect
              does this trace have to the resulting
              image (used for optimizing purposes)
       ior is the refractive index of the material where
              eye is in.

*/

const RGB Trace(Map *m, const XYZ &eye, const XYZ &dir, FLOAT_ARG effect, FLOAT_ARG ior)
{
    class recursion
    {
    public:
        recursion() { ++recdepth; }
        ~recursion() { --recdepth; }
    } recursion;
    
    const HitRec rec = m->FindObstacle(eye, dir);
    if(!rec.washit)
    {
    #if 0
        printf("No obstacles from ");
        eye.print();
        printf(" to ");
        dir.print();
        printf("\n");
    #endif
        /* Nothing there */
        return RGB(0,0,0);
    }
    
    const XYZ &hit      = rec.point;
    const XYZ &normal   = rec.normal;
    const Pigment &pigm = rec.Pigm;
    
#if 0
    printf("Hit at "); hit.print(); printf("; normal: "); normal.print(); printf("\n");
#endif

    RGB tmp(0.05, 0.05, 0.05);
    
    if(pigm.diffmul != 0.0)
    {
        Lights::const_iterator i;
        for(i=lights.begin(); i!=lights.end(); ++i)
        {
            XYZ lightdir = i->pos - hit;
            FLOAT lightdist = lightdir.size();
            lightdir /= lightdist;
            
        #if 1
            /* Shadow test without this object */
            const HitRec t = m->FindObstacle(hit+lightdir*1E-2, lightdir);
            //const HitRec t = m->FindObstacle(hit, lightdir, rec.obj);
            if(t.washit)
            {
                if(t.point.dist(hit) <= lightdist)
                {
                    /* It is shadowing. */
                    /* TODO: Find out how much!
                     * (The light is not a point but a ball)
                     */
                    
                    continue;
                }
            }
            /*
              TODO: Add light from all reflecting surfaces
                    (clipped by stuff in front of them).
            */
            
        #endif
            
            FLOAT diffect = normal.DotProduct(lightdir);
            //diffect=-diffect;
            if(diffect <= 0)continue;
            
            tmp += i->type * diffect * pigm.diffmul;
        #if 0
            printf("Light! Type: ");
            i->type.print();
            printf(" Effect: %.1"PF, diffect);
            printf("; pigment: ");
            pigm.print();
            printf(" (diffmul=%.1"PF, pigm.diffmul);
            printf(")\n");
        #endif
        }
        
        /*
        FLOAT neweffect = effect * pigm.diffmul * pigm.luminance() * 0.2;
        if(recdepth < 20)
        for(unsigned a=0; a<5; ++a)
        {
            XYZ tmpdir;
            tmpdir.x = drand48()*2.0-1.0;
            tmpdir.y = drand48()*2.0-1.0;
            tmpdir.z = drand48()*2.0-1.0;
            if(tmpdir.size() <= EPSILON)tmpdir.x = 1;
            tmpdir.Normalize();
            tmp += Trace(m, hit, tmpdir, neweffect, ior) * pigm.diffmul;
        }*/
    }
    //tmp *= pigm;
    
    if(pigm.specmul != 0.0)
    {
        XYZ rev = -dir;
        XYZ mirrored = rev.MirroredAround(normal);
        
        FLOAT neweffect = effect * pigm.specmul * pigm.luminance();
        /* Calculate how much this subtrace would affect to the pixel at max */
        /* Don't subtrace unless it may affect more than 1% */
        
        if(neweffect > 6)
        {
    #if 0
            fprintf(stderr, "Bright light! Bright light! effect=%.3"PF", specmul=%.3"PF", pigment=",
                    effect, pigm.specmul); fflush(stderr);
            pigm.print(); fflush(stdout);
            fprintf(stderr, "\n");
    #endif
            return tmp;
        }
        
        if(neweffect >= 0.004 && recdepth < 20)
        {
            // Reflected ray, same ior
            tmp += Trace(m, hit, mirrored, neweffect, ior) * pigm.specmul;
        }
    }
    
    if(pigm.penetration > EPSILON && recdepth < 20)
    {
        FLOAT localior      = rec.ior / ior;
        FLOAT criticalangle = asin(1.0 / localior);
        
        XYZ localnormal;
        FLOAT n = dir.DotProduct(normal);
        if(n <= 0.0)
        {
            localnormal = normal;
            n = -n;
        }
        else
            localnormal = -normal;
        
        FLOAT t = 1.0 + sqrt(localior) * (sqrt(n) - 1.0);
        if(t < 0.0)
        {
            /* Total internal reflection */
            XYZ rev = -dir;
            XYZ mirrored = rev.MirroredAround(normal);
            tmp += Trace(m, hit, mirrored, effect*pigm.penetration, localior) * pigm.penetration;
        }
        else
        {
            t = localior * n - sqrt(t);
            XYZ newdir = dir*ior + localnormal*t;
            tmp += Trace(m, hit, newdir, effect*pigm.penetration, localior) * pigm.penetration;
        }
    }
    
    tmp *= pigm;
    
    //tmp /= hit.dist(eye);
    return tmp;
}

void Trace(unsigned xres, unsigned yres,
           const XYZ &eye,
           const Matrix &meitriks,
           Map *m,
           const char *fn
          )
{
    static const unsigned char tgaBuf[] =
    {
        0,0,2, 0,0, 0,0, 0,  0,0, 0,0,
        xres&255, xres>>8,
        yres&255, yres>>8,
        24,32
    };
    
    FILE *fp = fopen(fn, "wb");
    fwrite(tgaBuf, 1, sizeof(tgaBuf), fp);
    
    /* refractive index of the environment eye is in */
    FLOAT ior = 1.000293;
        
    for(unsigned y=0; y<yres; ++y)
    {
        printf("%.1f%%\r", y*100.0/yres); fflush(stdout);
        
        const FLOAT ydisp = (y / (FLOAT)(yres-1) - 0.5);
        for(unsigned x=0; x<xres; ++x)
        {
            const FLOAT xdisp = (x / (FLOAT)(xres-1) - 0.5) * (xres/(FLOAT)yres);
            
            XYZ dir(xdisp, ydisp, 0.8);
            
            dir.Normalize();
            meitriks.Rotate(dir);
            
            RGB res = Trace(m, eye, dir, 1.0, ior);
            
            unsigned char R=res.getR(),
                          G=res.getG(),
                          B=res.getB();
            putc(B, fp);
            putc(G, fp);
            putc(R, fp);
        }
    }
    fclose(fp);
}
